import multiprocessing
import pandas as pd
import gc
from functools import partial
import time
from .utility import check_history_exclusion,phecode_leaf_to_root,time_first_diagnosis,n_process_check

# Define the worker function
def process_chunk(chunk, phecode_lst, disease_pair_index, node_dict, phecode_set, history_dict, diagnosis_dict, 
                    n_diagnosis_dict, phecode_info_dict, min_interval_days, max_interval_days, 
                    min_icd_num, sex_value_dict):
    
    # Initialize result dictionaries for this chunk
    chunk_ineligible_disease_dict = {}
    chunk_eligible_withdate_dict = {}
    chunk_d1d2_temporl_pair_dict = {}
    chunk_d1d2_com_pair_dict = {}
    chunk_all_diagnosis_level = {}
    
    from itertools import combinations
    
    # Process each individual in the chunk
    for id_, sex in chunk:
        temp_dineligible_list = []
        temp_dpair_temporal_lst = []
        temp_dpair_com_lst = []
        temp_deligible_dict_withdate = {}
        
        # Get individual's data
        diagnosis_ = diagnosis_dict[id_]
        n_diagnosis_ = n_diagnosis_dict[id_]
        history_ = history_dict[id_]
        
        # Generate a new all diagnosed phecode list correspond to the specified level
        all_diag_ind = set([node_dict.get(x, x) for x in history_ + list(diagnosis_.keys())]).intersection(phecode_set)
        all_diag_ind = [x for x in all_diag_ind if sum([n_diagnosis_.get(leaf, 0) for leaf in phecode_info_dict[x]['leaf_list']]) >= min_icd_num]
        chunk_all_diagnosis_level[id_] = all_diag_ind
        
        # Generate eligible disease dictionary
        for phecode in phecode_lst:
            leaf_lst = phecode_info_dict[phecode]['leaf_list']
            exl_lst = phecode_info_dict[phecode]['exclude_list']
            sex_specific = phecode_info_dict[phecode]['sex']
            
            if (check_history_exclusion(exl_lst, history_, n_diagnosis_, min_icd_num) == 0) and (sex_specific == 'Both' or sex_value_dict[sex_specific] == sex):
                date = time_first_diagnosis(leaf_lst, diagnosis_, n_diagnosis_, min_icd_num)
                if not pd.isna(date):
                    temp_deligible_dict_withdate[phecode] = date
            else:
                temp_dineligible_list.append(phecode)
        
        # Generate disease pair dictionary
        if len(temp_deligible_dict_withdate) <= 1:
            chunk_ineligible_disease_dict[id_] = temp_dineligible_list
            chunk_d1d2_temporl_pair_dict[id_] = temp_dpair_temporal_lst
            chunk_d1d2_com_pair_dict[id_] = temp_dpair_com_lst
            chunk_eligible_withdate_dict[id_] = temp_deligible_dict_withdate
        else:
            for d1, d2 in combinations(temp_deligible_dict_withdate, 2):
                date1, date2 = temp_deligible_dict_withdate[d1], temp_deligible_dict_withdate[d2]
                if abs((date1 - date2).days) > max_interval_days:
                    continue
                elif abs((date1 - date2).days) <= min_interval_days:
                    temp_dpair_com_lst.append(disease_pair_index[f'{d1}_{d2}'])  # order insensitive
                    continue
                else:
                    if date1 > date2:
                        temp_dpair_temporal_lst.append(disease_pair_index[f'{d2}_{d1}'])
                    else:
                        temp_dpair_temporal_lst.append(disease_pair_index[f'{d1}_{d2}'])
            
            # Save for the individual
            chunk_ineligible_disease_dict[id_] = temp_dineligible_list
            chunk_d1d2_temporl_pair_dict[id_] = temp_dpair_temporal_lst
            chunk_d1d2_com_pair_dict[id_] = temp_dpair_com_lst
            chunk_eligible_withdate_dict[id_] = temp_deligible_dict_withdate
    
    # Return this chunk's results
    return {
        'ineligible_disease': chunk_ineligible_disease_dict,
        'eligible_disease_withdate': chunk_eligible_withdate_dict,
        'd1d2_temporal_pair': chunk_d1d2_temporl_pair_dict,
        'd1d2_com_pair': chunk_d1d2_com_pair_dict,
        'all_diagnosis_level': chunk_all_diagnosis_level
    }

def parallel_d1d2_from_diagnosis_history(df:pd.DataFrame, id_col:str, sex_col:str, sex_value_dict:dict, 
                                phecode_lst:list, disease_pair_index:dict, history_dict:dict, diagnosis_dict:dict, n_diagnosis_dict:dict,
                                phecode_info_dict:dict, min_interval_days:int, max_interval_days:int, min_icd_num:int,
                                n_process:int) -> dict:
    """
    Parallel version of d1d2_from_diagnosis_history that constructs d1->d2 disease pairs
    for each individual from a list of significant phecodes using multiprocessing.
    
    Parameters
    ----------
        df : dataframe of phenotype data, contains at id and sex columns
        id_col : id column in the df
        sex_col : sex column in the df
        sex_value_dict: dictionary for coding 'Female' and 'Male' in the phenotype data
        phecode_lst : list of significant phecodes
        disease_pair_index : dictionary containing disease pair index
        history_dict : dictionary containing medical records history
        diagnosis_dict : dictionary containing diagnosis and date
        n_diagnosis_dict : number of phecode occurence
        phecode_info_dict : phecode information
        min_interval_days : minimum interval required for d1-d2 disease pair construction
        max_interval_days : maximum interval allowed for d1-d2 disease pair construction
        min_icd_num : The minimum number of ICD codes mapping to a specific phecode required for the phecode to be considered valid.
        n_process : Number of processes to use.

    Returns
    -------
        D1->D2 dictionary.
    """
    # Check multiprocessing parameters
    n_process,start_mehtod = n_process_check(n_process,'Disease pairs construction')
    
    # Generate a dictionary mapping leaf phecode to root phecode (shared across all individuals)
    node_dict = phecode_leaf_to_root(phecode_info_dict)
    phecode_set = set(phecode_info_dict.keys())
    
    # Extract IDs and sex values
    id_sex_pairs = df[[id_col, sex_col]].values
    chunk_size = int(len(id_sex_pairs)//n_process/2)
    
    # Divide the work into chunks
    chunks = [id_sex_pairs[i:i + chunk_size] for i in range(0, len(id_sex_pairs), chunk_size)]
    
    # Create a partial function with all the fixed parameters
    worker_func = partial(
        process_chunk,
        phecode_lst=phecode_lst,
        disease_pair_index=disease_pair_index,
        node_dict=node_dict,
        phecode_set=phecode_set,
        history_dict=history_dict,
        diagnosis_dict=diagnosis_dict,
        n_diagnosis_dict=n_diagnosis_dict,
        phecode_info_dict=phecode_info_dict,
        min_interval_days=min_interval_days,
        max_interval_days=max_interval_days,
        min_icd_num=min_icd_num,
        sex_value_dict=sex_value_dict
    )
    time_0 = time.time()
    # Set up multiprocessing pool with context manager
    with multiprocessing.get_context(start_mehtod).Pool(processes=n_process) as pool:
        # Map the worker function to the chunks
        results = pool.map(worker_func, chunks)
    time_1 = time.time()
    print(f'Parallel processing time: {time_1-time_0} seconds')
    # Merge results from all chunks
    ineligible_disease_dict = {}
    eligible_withdate_dict = {}
    d1d2_temporl_pair_dict = {}
    d1d2_com_pair_dict = {}
    all_diagnosis_level = {}
    
    while results:
        result = results.pop()  # removes and returns the last element
        ineligible_disease_dict.update(result['ineligible_disease'])
        eligible_withdate_dict.update(result['eligible_disease_withdate'])
        d1d2_temporl_pair_dict.update(result['d1d2_temporal_pair'])
        d1d2_com_pair_dict.update(result['d1d2_com_pair'])
        all_diagnosis_level.update(result['all_diagnosis_level'])
        gc.collect()
    
    # Final dictionary
    return {
        'disease_pair_index' : disease_pair_index,
        'ineligible_disease': ineligible_disease_dict,
        'eligible_disease_withdate': eligible_withdate_dict,
        'd1d2_temporal_pair': d1d2_temporl_pair_dict,
        'd1d2_com_pair': d1d2_com_pair_dict,
        'all_diagnosis_level': all_diagnosis_level
    }


from .analysis import (
    phewas,
    comorbidity_strength, 
    binomial_test, 
    comorbidity_network, 
    disease_trajectory)
from .data_management import DiseaseNetworkData
import numpy as np
import os
from .utility import n_process_check,threshold_check,validate_method_specific_kwargs,matching_var_check
from .utility import correction_method_check
import time
import pandas as pd

def disease_network_pipeline(
    data:DiseaseNetworkData, 
    n_process:int, 
    n_threshold_phewas:int, 
    n_threshold_comorbidity:int,
    output_dir:str,
    project_prefix:str,
    keep_positive_associations:bool=False,
    save_intermediate_data:bool=False,
    system_exl:list=None, 
    pipeline_mode:str='v1',
    method:str='RPCN',
    covariates:list=None,
    matching_var_dict:dict={'sex':'exact'},
    matching_n:int=2,
    min_interval_days:int=0,
    max_interval_days:int=np.inf,
    enforce_temporal_order:bool=False,
    correction:str='bonferroni',
    cutoff=0.05,
    **kwargs) -> pd.DataFrame:
    """
    This is a pipeline function to run the whole disease network analysis.

    Parameters
    ----------
    data : DiseaseNetworkData
        The DiseaseNetworkData object.

    n_process : int
        Specifies the number of parallel processes to use for the disease network analysis.
        Multiprocessing is enabled when `n_process` is set to a value greater than one.

    n_threshold_phewas : int
        The minimum number of cases within the exposed group required for a phecode to be included in the PheWAS analysis.
        This parameter will be passed to the phewas function.
        See the phewas function for more information.

    n_threshold_comorbidity : int
        The minimum number of individuals in the exposed group in which a disease pair must co-occur (temporal or non-temporal) to be included in the comorbidity strength estimation.
        This parameter will be passed to the comorbidity_strength function.
        See the comorbidity_strength function for more information.
    
    output_dir : str
        Directory path to store output files generated by the pipeline.

    project_prefix : str
        Prefix for naming output files and intermediate data.

    keep_positive_associations : bool, default=False
        If True, retains only diseases with hazard ratio (HR) > 1 from the PheWAS analysis,
        and disease pairs with positive correlations from the comorbidity strength estimation.

    save_intermediate_data : bool, default=False
        If True, intermediate DiseaseNetworkData objects created by the `DiNetxify.DiseaseNetworkData.disease_pair` function are saved to disk.
    
    system_exl : list, default=None
        List of phecode systems to exclude from the analysis. 
        List of eligible phecode systems: 
            circulatory system; congenital anomalies; dermatologic; digestive; 
            endocrine/metabolic; genitourinary; hematopoietic; infectious diseases; injuries & poisonings; 
            mental disorders; musculoskeletal; neoplasms; neurological; pregnancy complications; 
            respiratory; sense organs; symptoms; others.
        This parameter will be passed to the phewas function.
        See the phewas function for more information.
    
    pipeline_mode : str, default='v1'
        Specifies the analysis order. Available modes:
            - 'v1': PheWAS → comorbidity strength → binomial test → (comorbidity network analysis/disease trajectory analysis)
            - 'v2': PheWAS → comorbidity strength → comorbidity network analysis → binomial test → disease trajectory analysis
        In 'v1', the binomial test does not depend on results from the comorbidity network analysis; thus, disease trajectory and comorbidity network analyses can be conducted independently.
        In 'v2', the binomial test is performed only on disease pairs identified as significant by the comorbidity network analysis, making the disease trajectory analysis dependent on these results.

    method : str, default='RPCN'
        The method to use for the comorbidity network and disease trajectory analysis.
        Available methods are:
            - 'RPCN: Regularized Partial Correlation Network.
            - 'PCN_PCA: Partial Correlation Network with PCA.
            - 'CN': Correlation Network.
        This parameter will be passed to the comorbidity_network and disease_trajectory function.
        See the these two functions for more information.
    
    covariates : list, default=None
        List of covariates to adjust for in the PheWAS, comorbidity network and disease trajectory analysis.
        To include the required variable sex as a covariate, always use 'sex' instead of its original column name.
        For other covariates specified in the 'DiNetxify.DiseaseNetworkData.phenotype_data()' function, use their original column names.
        This parameter will be passed to the phewas, comorbidity_network and disease_trajectory function.
        See these three functions for more information.
    
    matching_var_dict : dict, default={'sex':'exact'}
        Specifies the matching variables and the criteria used for incidence density sampling.
        For categorical and binary variables, the matching criteria should always be 'exact'.
        For continuous variables, provide a scalar greater than 0 as the matching criterion, indicating the maximum allowed difference when matching.
        To include the required variable sex as a matching variable, always use 'sex' instead of its original column name.
        This parameter will be passed to the disease_trajectory function.
        See the disease_trajectory function for more information.
    
    matching_n : int, default=2
        Specifies the maximum number of matched controls for each case.
        This parameter will be passed to the disease_trajectory function.
        See the disease_trajectory function for more information.
    
    min_interval_days : int/float, default=0
        Minimum required time interval (in days) between diagnosis dates when constructing temporal D1 → D2 disease pair for each individual.
        This parameter will be passed to the DiNetxify.DiseaseNetworkData.disease_pair function.
        See the disease_pair function for more information.

    max_interval_days : int/float, default=np.inf
        Maximum allowed time interval (in days) between diagnosis dates when constructing temporal and non-temporal D1-D2 disease pair for each individual.
        This parameter will be passed to the DiNetxify.DiseaseNetworkData.disease_pair function.
        See the disease_pair function for more information.
    
    enforce_temporal_order : bool, default=False
        If set to True, exclude individuals with non-temporal D1-D2 pair when performing the binomial test;
        also applies the specified minimum and maximum time intervals when performing disease trajectory analysis.
        See 'enforce_temporal_order' parameter in binomial_test function and 'enforce_time_interval' parameter in disease_trajectory function.
    
    correction : str, default='bonferroni'
        Method for p-value correction from the statsmodels.stats.multitest.multipletests.
        Available methods are:
        none : no correction
        bonferroni : one-step correction
        sidak : one-step correction
        holm-sidak : step down method using Sidak adjustments
        holm : step-down method using Bonferroni adjustments
        simes-hochberg : step-up method (independent)
        hommel : closed method based on Simes tests (non-negative)
        fdr_bh : Benjamini/Hochberg (non-negative)
        fdr_by : Benjamini/Yekutieli (negative)
        fdr_tsbh : two stage fdr correction (non-negative)
        fdr_tsbky : two stage fdr correction (non-negative)
        See https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html for more details.
    
    cutoff : float, default=0.05
        The significance threshold for adjusted p-values.
    
    **kwargs
        RPCN Method Parameters:
            alpha : non-negative scalar
                The weight multiplying the l1 penalty term for other diseases covariates. 
                Ignored if 'auto_penalty' is enabled.
            auto_penalty : bool, default=True
                If 'True', automatically determines the best 'alpha' based on model AIC value.
            alpha_range : tuple, default=(1,15)
                When 'auto_penalty' is True, search the optimal 'alpha' in this range.
            scaling_factor : positive scalar, default=1
                The scaling factor for the alpha when 'auto_penalty' is True.

        PCN_PCA Method Parameters:
            n_PC : int, default=5
                Fixed number of principal components to include in each model.
            explained_variance : float
                Cumulative explained variance threshold to determine the number of principal components. 
                Overrides 'n_PC' if specified.
    
    Returns
    -------
    tuple
        A tuple of DataFrames for result of PheWAS, comorbidity strength estimation, comorbidity network analysis, binomial test, and disease trajectory analysis.
    """
    
    # --------check each parameter--------
    #data type check
    if not isinstance(data,DiseaseNetworkData):
        raise TypeError("The input 'data' must be a DiseaseNetworkData object.")
    #attribute check
    data_attrs = ['phenotype_df', 'diagnosis', 'n_diagnosis', 'history']
    for attr in data_attrs:
        if getattr(data, attr) is None:
            raise ValueError(f"Attribute '{attr}' is empty.")
    #n_process check
    n_process, _ = n_process_check(n_process,'PheWAS')
    #check threshold
    n_exposed = data.get_attribute('phenotype_statistics')['n_exposed']
    n_threshold_phewas = threshold_check(None,n_threshold_phewas,n_exposed)
    n_threshold_comorbidity = threshold_check(None,n_threshold_comorbidity,n_exposed)
    if n_threshold_comorbidity > n_threshold_phewas:
        raise ValueError("The 'n_threshold_comorbidity' must be less than or equal to 'n_threshold_phewas'.")
    #output_dir check
    if not isinstance(output_dir,str):
        raise TypeError("The provided input 'output_dir' must be a string.")
    if not os.path.exists(output_dir):
        raise ValueError("The provided output directory does not exist.")
    #project_prefix check
    if not isinstance(project_prefix,str):
        raise TypeError("The provided input 'project_prefix' must be a string.")
    #keep_positive_associations check
    if not isinstance(keep_positive_associations,bool):
        raise TypeError("The provided input 'keep_positive_associations' must be a bool.")
    #save_intermediate_data check
    if not isinstance(save_intermediate_data,bool):
        raise TypeError("The provided input 'save_intermediate_data' must be a bool.")
    #check system_exl
    # phecode_info = data.phecode_info
    # phecode_lst_all = filter_phecodes(phecode_info, None, system_exl, None, None)
    # print(f'A total of {len(phecode_lst_all)} phecodes included in the PheWAS analysis.')
    #pipeline_mode check
    if pipeline_mode not in ['v1','v2']:
        raise ValueError("Invalid pipeline_mode. Must be either 'v1' or 'v2'.")
    #method and kwargs check
    if kwargs:
        parameter_dict = validate_method_specific_kwargs(method,**kwargs)
        if kwargs:
            raise ValueError(f"Unexpected keyword arguments provided: {kwargs.keys()}")
    else:
        parameter_dict = {}
    #check covariates
    # covariates = covariates_check(covariates,data.get_attribute('phenotype_info'))
    #check matching_var_dict
    matching_var_check(matching_var_dict,data.get_attribute('phenotype_info'))
    #check matching_n
    if not isinstance(matching_n,int) or matching_n<1:
        raise ValueError("The input 'matching_n' must be a positive integer.")
    #check min and max interval 
    if not isinstance(min_interval_days, (float,int)):
        raise TypeError("The provided input 'min_interval_days' must be a int")
    if min_interval_days<0:
        raise ValueError("The provided input 'min_interval_days' is not valide.")
    if not isinstance(max_interval_days, (float,int)):
        raise TypeError("The provided input 'max_interval_days' must be a int")
    if max_interval_days<0 or max_interval_days<=min_interval_days:
        raise ValueError("The provided input 'max_interval_days' is not valide.")
    #check enforce_temporal_order
    if not isinstance(enforce_temporal_order,bool):
        raise TypeError("The provided input 'enforce_temporal_order' must be a bool.")
    #check correction
    correction_method_check(correction,cutoff)

    # time used
    time_start = time.time()

    # --------run the pipeline--------
    #run PheWAS
    phewas_log_file = os.path.join(output_dir,f'{project_prefix}_phewas.log')
    phewas_result_file = os.path.join(output_dir,f'{project_prefix}_phewas_result.csv')
    phewas_result = phewas(data=data, n_threshold=n_threshold_phewas,
                            n_process=n_process, system_exl=system_exl, covariates=covariates, 
                            correction=correction, cutoff=cutoff, log_file=phewas_log_file)
    phewas_result.to_csv(phewas_result_file,index=False)
    #filter the positive associations if needed
    if keep_positive_associations:
        if data.study_design == "exposed-only cohort":
            pass
        else:
            phewas_result = phewas_result[phewas_result['phewas_coef']>=0]
    #number of significant phecodes, saved it to the results_dict
    n_sig_phecodes = len(phewas_result[phewas_result['phewas_p_significance']==True])
    print(f'Number of significant diseases identified in the PheWAS analysis: {n_sig_phecodes}.')
    #generate intermediate data
    data.disease_pair(phewas_result=phewas_result,
                      min_interval_days=min_interval_days,
                      max_interval_days=max_interval_days,
                      force=True)
    if save_intermediate_data:
        data_intermediate_file = os.path.join(output_dir,f'{project_prefix}_data_withtra')
        data.save(data_intermediate_file)
    
    #run comorbidity strength estimation
    comorbidity_log_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_strength.log')
    com_strength_result_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_strength_result.csv')
    com_strength_result = comorbidity_strength(data=data, n_threshold=n_threshold_comorbidity,n_process=n_process, 
                                               correction_phi=correction, correction_RR=correction,
                                               cutoff_phi=cutoff, cutoff_RR=cutoff,
                                               log_file=comorbidity_log_file)
    com_strength_result.to_csv(com_strength_result_file,index=False)
    if keep_positive_associations:
        com_strength_result = com_strength_result[(com_strength_result['phi'] > 0) & (com_strength_result['RR'] > 1)]
    #number of significant comorbidity pairs, saved it to the results_dict
    n_sig_com_strength = len(com_strength_result[(com_strength_result['RR_p_significance']==True) & 
                                                (com_strength_result['phi_p_significance']==True)])
    print(f'Number of significant disease pairs identified in the comorbidity strength estimation: {n_sig_com_strength}.')
    
    #mode v1
    if pipeline_mode == 'v1':
        #run binomial test
        binomial_log_file = os.path.join(output_dir,f'{project_prefix}_binomial.log')
        binomial_result_file = os.path.join(output_dir,f'{project_prefix}_binomial_result.csv')
        binomial_result = binomial_test(data=data, comorbidity_strength_result=com_strength_result,
                                        n_process=1, correction=correction, cutoff=cutoff,
                                        enforce_temporal_order=enforce_temporal_order,
                                        log_file=binomial_log_file)
        binomial_result.to_csv(binomial_result_file,index=False)
        #number of temporal pairs
        n_temporal_pairs = len(binomial_result[binomial_result['binomial_p_significance']==True])
        print(f'Number of significant disease pairs identified in the binomial test: {n_temporal_pairs}.')
        #run comorbidity network analysis
        com_network_log_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_network.log')
        com_network_result_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_network_result.csv')
        if parameter_dict:
            com_network_result = comorbidity_network(data=data, comorbidity_strength_result=com_strength_result,
                                                    n_process=n_process, method=method, 
                                                    covariates=covariates, correction=correction, cutoff=cutoff,
                                                    log_file=com_network_log_file, **parameter_dict)
        else:
            com_network_result = comorbidity_network(data=data, comorbidity_strength_result=com_strength_result,
                                                    n_process=n_process, method=method, 
                                                    covariates=covariates, correction=correction, cutoff=cutoff,
                                                    log_file=com_network_log_file)
            
        com_network_result.to_csv(com_network_result_file,index=False)
        #number of significant comorbidity network pairs
        n_sig_com_network = len(com_network_result[com_network_result['comorbidity_p_significance']==True])
        print(f'Number of significant disease pairs identified in the comorbidity network analysis: {n_sig_com_network}.')
    elif pipeline_mode == 'v2':
        #run comorbidity network analysis
        com_network_log_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_network.log')
        com_network_result_file = os.path.join(output_dir,f'{project_prefix}_comorbidity_network_result.csv')
        if parameter_dict:
            com_network_result = comorbidity_network(data=data, comorbidity_strength_result=com_strength_result,
                                                    n_process=n_process, method=method, 
                                                    covariates=covariates, correction=correction, cutoff=cutoff,
                                                    log_file=com_network_log_file, **parameter_dict)
        else:
            com_network_result = comorbidity_network(data=data, comorbidity_strength_result=com_strength_result,
                                                    n_process=n_process, method=method, 
                                                    covariates=covariates, correction=correction, cutoff=cutoff,
                                                    log_file=com_network_log_file)
        com_network_result.to_csv(com_network_result_file,index=False)
        #number of significant comorbidity network pairs
        n_sig_com_network = len(com_network_result[com_network_result['comorbidity_p_significance']==True])
        print(f'Number of significant disease pairs identified in the comorbidity network analysis: {n_sig_com_network}.')
        #run binomial test
        binomial_log_file = os.path.join(output_dir,f'{project_prefix}_binomial.log')
        binomial_result_file = os.path.join(output_dir,f'{project_prefix}_binomial_result.csv')
        binomial_result = binomial_test(data=data, comorbidity_strength_result=com_strength_result,
                                        comorbidity_network_result=com_network_result,
                                        n_process=1, correction=correction, cutoff=cutoff,
                                        enforce_temporal_order=enforce_temporal_order,
                                        log_file=binomial_log_file)
        binomial_result.to_csv(binomial_result_file,index=False)
        #number of temporal pairs
        n_temporal_pairs = len(binomial_result[binomial_result['binomial_p_significance']==True])
        print(f'Number of significant disease pairs identified in the binomial test: {n_temporal_pairs}.')

    #run disease trajectory analysis
    # first get the final covariates list for disease trajectory analysis, 
    # by exluding potential categorical variables used for matching
    # covariates_trajectory = covariates_check(covariates,data.get_attribute('phenotype_info'),matching_var_dict,exclude=True)
    #analysis
    trajectory_log_file = os.path.join(output_dir,f'{project_prefix}_trajectory.log')
    trajectory_result_file = os.path.join(output_dir,f'{project_prefix}_trajectory_result.csv')
    if parameter_dict:
        trajectory_result = disease_trajectory(data=data, comorbidity_strength_result=com_strength_result,
                                            binomial_test_result=binomial_result,
                                            n_process=n_process, method=method, 
                                            covariates=covariates,
                                            matching_var_dict=matching_var_dict, matching_n=matching_n,
                                            correction=correction, cutoff=cutoff,
                                            log_file=trajectory_log_file, **parameter_dict)
    else:
        trajectory_result = disease_trajectory(data=data, comorbidity_strength_result=com_strength_result,
                                            binomial_test_result=binomial_result,
                                            n_process=n_process, method=method, 
                                            covariates=covariates,
                                            matching_var_dict=matching_var_dict, matching_n=matching_n,
                                            correction=correction, cutoff=cutoff,
                                            log_file=trajectory_log_file)
        
    trajectory_result.to_csv(trajectory_result_file,index=False)
    #number of significant disease trajectories
    n_sig_trajectory = len(trajectory_result[trajectory_result['trajectory_p_significance']==True])
    print(f'Number of significant disease pairs identified in the disease trajectory analysis: {n_sig_trajectory}.')

    #calculate time used
    time_end = time.time()
    time_used = round((time_end-time_start)/60,2)
    print(f'The disease network analysis pipeline has been completed, total time: {time_used} mins.')
    
    #return all the results df in order
    return phewas_result, com_strength_result, com_network_result, binomial_result, trajectory_result

    



import pandas as pd
import numpy as np
import scipy.stats as stats
import warnings
warnings.filterwarnings('ignore')

def categorical(df,var,group_var,group_var_value,table1):
    """
    Function to generate the table rows for categorical variables
    """
    var_values = df[var].value_counts().index
    all_rows = []
    #a header row for the variable
    all_rows.append([f'{var} (n, %)']+['']*len(group_var_value)+[''])
    #a row for each value of the variable
    for value in var_values:
        row = [f'{var}={value}']
        for group in group_var_value:
            group_data = df[df[group_var]==group]
            numbers = group_data[df[var]==value].shape[0]
            percentage = numbers/group_data.shape[0]*100
            row.append(f'{numbers:,} ({percentage:.2f}%)')
        if value == var_values[-1]:
            if len(group_var_value) >= 2:
                cont_table = pd.crosstab(df[var],df[group_var])
                try:
                    chitest = stats.chi2_contingency(cont_table)
                    p_value = f"Chi-squared test p-value={chitest[1]:.3e}"
                except:
                    chitest = stats.fisher_exact(cont_table)
                    p_value = f"Fisher's exact test p-value={chitest[1]:.3e}"
                row.append(p_value)
            else:
                row.append('NA')
        else:
            row.append('')
        all_rows.append(row)
    table1.extend(all_rows)

def continuous_normal(df,var,group_var,group_var_value,table1):
    """
    Function to generate the table rows for normally distributted continuous variables
    """
    row_continous = [f'{var} (mean, SD)']
    for group in group_var_value:
        group_data = df[df[group_var]==group]
        #using f-string to format the number
        mean_sd = f'{group_data[var].mean():,.2f} ({group_data[var].std():,.2f})'
        row_continous.append(mean_sd)
    if len(group_var_value) == 2:
        ttest = stats.ttest_ind(df[df[group_var]==group_var_value[0]][var],df[df[group_var]==group_var_value[1]][var])
        p_value = f"Student's T-test p-value={ttest[1]:.3e}"
        row_continous.append(p_value)
    else: 
        row_continous.append('NA')
    table1.append(row_continous)

def continuous_unnormal(df,var,group_var,group_var_value,table1):
    """
    Function to generate the table rows for unnormally distributted continuous variables
    """
    row_continous = [f'{var} (median, IQR)']
    for group in group_var_value:
        group_data = df[df[group_var]==group]
        #using f-string to format the number
        median_iqr = f'{group_data[var].median():,.2f} ({group_data[var].quantile(0.25):,.2f}-{group_data[var].quantile(0.75):,.2f})'
        row_continous.append(median_iqr)
    if len(group_var_value) == 2:
        utest = stats.mannwhitneyu(df[df[group_var]==group_var_value[0]][var],df[df[group_var]==group_var_value[1]][var])
        p_value = f'Mann-Whitney U test p-value={utest[1]:.3e}'
        row_continous.append(p_value)
    else:
        row_continous.append('NA')
    table1.append(row_continous)

def desceibe_table(df:pd.DataFrame,var_lst:list,var_type_lst:list,group_var:str,
                   sex_value_dict:dict, continuous_stat_mode:str='auto',group_var_value:list=[1,0]):
    """
    Function to generate the simple table1

    df: pandas DataFrame
    var_lst: list of variable names
    var_type_lst: list of variable types
    group_var: the group variable
    sex_value_dict: mapping dictionary for sex variable
    continuous_stat_mode: either 'auto', 'normal' or 'nonnormal'
    group_var_value: the value of group variable
    """
    table1 = []
    #table 1 header
    header = ['Variable']
    for group in group_var_value:
        n_total = df[df[group_var]==group].shape[0]
        header.append(f"{group_var}={group} (n={n_total:,})")
    header.append('Test and p-value')
    #table1.append(header)
    #table 1 content
    for var,type_ in zip(var_lst,var_type_lst):
        if type_ == 'continuous':
            if continuous_stat_mode=='auto':
                p = stats.shapiro(df[var].values)[1]
                if p>0.05:
                    continuous_normal(df,var,group_var,group_var_value,table1)
                else:
                    continuous_unnormal(df,var,group_var,group_var_value,table1)
            elif continuous_stat_mode=='normal':
                continuous_normal(df,var,group_var,group_var_value,table1)
            elif continuous_stat_mode=='nonnormal':
                continuous_unnormal(df,var,group_var,group_var_value,table1)
        elif type_ == 'categorical' or type_ == 'binary':
            if var == 'sex':
                df[var] = df[var].map({j:i for i,j in sex_value_dict.items()})
            #fill the na with 'NA'
            df[var] = df[var].fillna('NAN')
            categorical(df,var,group_var,group_var_value,table1)
    return pd.DataFrame(table1,columns=header)




# -*- coding: utf-8 -*-
"""
Created on Sun Dec 15 02:13:51 2024

@author: Can Hou - Biomedical Big data center of West China Hospital, Sichuan University
"""

import pandas as pd
import numpy as np
from statsmodels.discrete.discrete_model import Logit
import time
import gc
from .utility import write_log, find_best_alpha_and_vars, check_variance_vif_single

import warnings
warnings.filterwarnings('ignore')

def logistic_model(args):
    """
    Fit a LR model to verify the comorbidity association between a disease pair.

    Parameters
    ----------
    d1 : str, phecode 1.
    d2 : str, phecode 2.

    Global Variables:
    ----------
    phenotype_df_exposed : pd.DataFrame, phenotypic data for exposed individuals only.
    trajectory_ineligible : dict, trajectory ineligible disease dictionary.
    trajectory_eligible_withdate : dict, trajectory eligible disease (with date) dictionary.
    all_diagnosis_level : list, list of all diagnosed phecodes, with phecode truncated to corresponding level
    covariates : list, list of covariates to be included in the model.
    all_diseases_lst : list, list of other diseases to be included.
    log_file : str, Path and prefix for the log file
    parameters : dict, other arguments, including method and the associated parameters.

    Returns
    -------
    Result list

    """
    #shared global data
    global phenotype_df_exposed_
    global id_col_
    global trajectory_ineligible_
    global trajectory_eligible_withdate_
    global all_diagnosis_level_
    global covariates_
    global all_diseases_lst_
    global log_file_
    global parameters_

    d1, d2 = args
    #default columns
    d1_col = 'd1'
    d2_col = 'd2'
    constant_col = 'constant'
    forcedin_vars = [d1_col,constant_col]

    #method and parameters
    method = parameters_['method']
    if method == 'RPCN':
        #alpha_initial = [1, 10, 20, 30, 40, 50] #alpha starting value range if using auto_penalty
        auto_penalty = parameters_['auto_penalty']
        alpha_single = parameters_['alpha']
        alpha_range = parameters_['alpha_range']
        scaling_factor = parameters_['scaling_factor']
    elif method == 'PCN_PCA':
        pca_number = parameters_.get('explained_variance',parameters_.get('n_PC')) #retrive explained_variance first if given, otherwise use n_PC
    
    #filtering the dataframe first
    N = len(phenotype_df_exposed_)
    d1d2_eligible_lst = [id_ for id_,vals in trajectory_ineligible_.items() if d1 not in vals and d2 not in vals]
    df_analysis = phenotype_df_exposed_[phenotype_df_exposed_[id_col_].isin(d1d2_eligible_lst)]
    
    #create other diseases variable (already generated in the main function)
    if method in ['RPCN','PCN_PCA']:
        diseases_lst_ = [x for x in all_diseases_lst_ if x!=d1 and x!=d2]
        all_diseases_var = [str(x) for x in diseases_lst_]
    
    #d1 and d2 variable
    df_analysis[d1_col] = df_analysis[id_col_].apply(lambda x: 1 if d1 in trajectory_eligible_withdate_[x] else 0)
    df_analysis[d2_col] = df_analysis[id_col_].apply(lambda x: 1 if d2 in trajectory_eligible_withdate_[x] else 0)
    df_analysis[constant_col] = 1
    #statistics
    n = len(df_analysis) #number of individuals in the matched case-control study
    N_d2 = len(df_analysis[df_analysis[d2_col]==1])
    N_nod2 = len(df_analysis[df_analysis[d2_col]==0])
    N_d1_withd2 = len(df_analysis[(df_analysis[d2_col]==1) & (df_analysis[d1_col]==1)])
    N_d1_nod2 = len(df_analysis[(df_analysis[d2_col]==0) & (df_analysis[d1_col]==1)])

    #result list
    result_lst = [d1,d2,f'{d1}-{d2}',N,n,f'{N_d1_withd2}/{N_d2}',f'{N_d1_nod2}/{N_nod2}']
    
    #time and message
    time_start = time.time()
    message = f'{d1} and {d2}: '

    #check phnotypic covariates (applied to all methods)
    del_covariates = check_variance_vif_single(df_analysis,
                                               forcedin_vars,covariates_,
                                               vif_cutoff='phenotypic_covar')
    final_covariates= [x for x in covariates_ if x not in del_covariates]
    
    #simple method
    if method == 'CN':
        try:
            #list of final variables to be included
            final_model_vars = forcedin_vars+final_covariates
            model = Logit(np.asarray(df_analysis[d2_col], dtype=int),
                          np.asarray(df_analysis[final_model_vars], dtype=float))
            result_final = model.fit(disp=False, method='bfgs')
            beta,se,p,aic = result_final.params[0], result_final.bse[0],result_final.pvalues[0],result_final.aic
            zvalue_dict = {var:z for var,z in zip(final_model_vars,result_final.tvalues)}
            result_lst += [method,f'fitted and delete the covariate(s): {del_covariates}',
                            f'{final_model_vars}',f'{zvalue_dict}',beta,se,p,aic]
            message += f'method={method}; successfully fitted; '
        except Exception as e:
            message += f'method={method}; error encountered: {e}; '
            result_lst += [method,str(e)]
    
    #partial correlation method
    elif method == 'RPCN':
        #check disease variables
        del_diseases_var = check_variance_vif_single(df_analysis,
                                                     forcedin_vars,all_diseases_var,
                                                     vif_cutoff='disease_covar')
        all_diseases_var = [x for x in all_diseases_var if x not in del_diseases_var]
        #generate alpha lst after removing these variables
        alpha_lst = np.array([0]*len(forcedin_vars) + [1]*len(all_diseases_var))
        #variables for first model for selecting disease variables
        model_1_vars = forcedin_vars+all_diseases_var
        #method 1
        if auto_penalty:
            alpha_lst = alpha_lst * scaling_factor #scaling_factor only for auto_penalty
            try:
                #model
                model = Logit(np.asarray(df_analysis[d2_col], dtype=int),
                              np.asarray(df_analysis[model_1_vars], dtype=float))
                #search within the defined range, get the diseases variables that will be included plus forced-in variables
                final_best_alpha, final_disease_forcedin_vars = find_best_alpha_and_vars(model,alpha_range,alpha_lst,model_1_vars)
                final_disease_vars = [x for x in final_disease_forcedin_vars if x not in forcedin_vars]
                #fit the final model
                final_model_vars = final_disease_forcedin_vars+final_covariates
                #restrict the dataset to the final model variables
                df_analysis = df_analysis[final_model_vars+[d2_col]]
                model_final = Logit(np.asarray(df_analysis[d2_col],dtype=int),
                                    np.asarray(df_analysis[final_model_vars],dtype=float))
                result_final = model_final.fit(disp=False, method='bfgs')
                beta,se,p,aic = result_final.params[0],result_final.bse[0],result_final.pvalues[0],result_final.aic
                #get the z-value dictionary
                zvalue_dict = {var:z for var,z in zip(final_model_vars,result_final.tvalues)}
                result_lst += [f'{method}_auto',f'fitted and delete the diseases variable(s): {del_diseases_var} and covariate(s): {del_covariates}',
                               f'{final_model_vars}',f'{zvalue_dict}',final_best_alpha,beta,se,p,aic]
                message += f'method={method}_auto (alpha={final_best_alpha}, number of other disease included as covariates: {len(final_disease_vars)}); successfully fitted; '
            except Exception as e:
                result_lst += [f'{method}_auto', str(e)]
                message += f'method={method}_auto; error encountered: {e}; '

        else:
            try:
                #fit the initial model to get the non-zero disease list
                model = Logit(np.asarray(df_analysis[d2_col],dtype=int),
                              np.asarray(df_analysis[model_1_vars],dtype=float))
                result = model.fit_regularized(method='l1', alpha=alpha_lst*alpha_single, disp=False)
                #get list of variables with none zero coef
                non_zero_indices = np.nonzero(result.params != 0)[0]
                final_disease_forcedin_vars = [model_1_vars[i] for i in non_zero_indices]
                final_disease_vars = [x for x in final_disease_forcedin_vars if x not in forcedin_vars]
                #fit the final model
                final_model_vars = final_disease_forcedin_vars+final_covariates
                #restrict the dataset to the final model variables
                df_analysis = df_analysis[final_model_vars+[d2_col]]
                model_final = Logit(np.asarray(df_analysis[d2_col],dtype=int),
                                    np.asarray(df_analysis[final_model_vars]),dtype=float)
                result_final = model_final.fit(disp=False,method='bfgs')
                beta,se,p,aic = result_final.params[0], result_final.bse[0],result_final.pvalues[0],result_final.aic
                #get the z-value dictionary
                zvalue_dict = {var:z for var,z in zip(final_model_vars,result_final.tvalues)}
                result_lst += [f'{method}_fixed_alpha',f'fitted and delete the diseases variable(s): {del_diseases_var} and covariate(s): {del_covariates}',
                               f'{final_model_vars}',f'{zvalue_dict}',alpha_single,beta,se,p,aic]
                message += f'method={method}_fixed_alpha (alpha={alpha_single}, number of other disease included as covariates: {len(final_disease_vars)}); successfully fitted; '
            except Exception as e:
                result_lst += [f'{method}_fixed_alpha', str(e)]
                message += f'method={method}_fixed_alpha (alpha={alpha_single}); error encountered: {e}; '

    elif method == 'PCN_PCA':
        from sklearn.decomposition import PCA # type: ignore
        try:
            #generate PC from other diseases variables
            pca = PCA(n_components=pca_number)
            disease_vars_transformed = pca.fit_transform(np.asarray(df_analysis[all_diseases_var],dtype=int))
            #delete the original disease variables
            df_analysis = df_analysis.drop(all_diseases_var,axis=1)
            all_pca_vars = [f'PCA_{i}' for i in range(disease_vars_transformed.shape[1])]
            disease_vars_transformed = pd.DataFrame(disease_vars_transformed,columns=all_pca_vars)
            disease_vars_transformed.index = df_analysis.index
            df_analysis = pd.concat([df_analysis,disease_vars_transformed],axis=1)
            #delete the disease_vars_transformed
            del disease_vars_transformed
            gc.collect()
            variance_explained = sum(pca.explained_variance_ratio_)
            #check the VIF of PCA variables
            del_pca_var = check_variance_vif_single(df_analysis,
                                                    forcedin_vars,all_pca_vars,
                                                    vif_cutoff='pca_covar')
            final_pca_var = [x for x in all_pca_vars if x not in del_pca_var]
            #fit the final model
            final_model_vars = forcedin_vars+final_pca_var+final_covariates
            model_final = Logit(np.asarray(df_analysis[d2_col],dtype=int),
                                np.asarray(df_analysis[final_model_vars],dtype=float))
            result_final = model_final.fit(disp=False,method='bfgs')
            beta,se,p,aic = result_final.params[0], result_final.bse[0],result_final.pvalues[0],result_final.aic
            zvalue_dict = {var:z for var,z in zip(final_model_vars,result_final.tvalues)}
            result_lst += [f'{method}_n_components={pca_number}',f'fitted and delete the PC variable(s): {del_pca_var} and covariate(s): {del_covariates}',
                           f'{final_model_vars}',f'{zvalue_dict}',variance_explained,beta,se,p,aic]
            message += f'method={method}_n_components={pca_number} (number of PC included as covariates: {len(final_pca_var)}, total variance explained by PC: {variance_explained:.3f}); successfully fitted; '
        except Exception as e:
            result_lst += [f'{method}_n_components={pca_number}',str(e)]
            message += f'method={method}_n_components={pca_number}; error encountered: {e}; '
    
    #print and return
    time_end = time.time()
    time_spend = time_end - time_start
    message += f'(elapsed {time_spend:.2f}s)\n'
    write_log(log_file_,message)
    return result_lst

def logistic_model_wrapper(d1:float,d2:float,phenotype_df_exposed:pd.DataFrame,id_col,trajectory_ineligible:dict,
                            trajectory_eligible_withdate:dict,all_diagnosis_level:dict,covariates:list,
                            all_diseases_lst:list,log_file:str,parameters:dict):
    """
    Wrapper for logistic_model that assigns default values to global variables if needed.

    Parameters
    ----------
    d1 : float, phecode 1.
    d2 : float, phecode 2.
    phenotype_df_exposed : pd.DataFrame, phenotypic data for exposed individuals only.
    trajectory_ineligible : dict, trajectory ineligible disease dictionary.
    trajectory_eligible_withdate : dict, trajectory eligible disease (with date) dictionary.
    all_diagnosis_level : list, list of all diagnosed phecodes, with phecode truncated to corresponding level
    covariates : list, list of covariates to be included in the model.
    all_diseases_lst : list, list of other diseases to be included.
    log_file : str, Path and prefix for the log file
    parameters : dict, other arguments, including method and the associated parameters.

    Returns
    -------
    Result list

    """
    #shared global data
    global phenotype_df_exposed_
    global id_col_
    global trajectory_ineligible_
    global trajectory_eligible_withdate_
    global all_diagnosis_level_
    global covariates_
    global all_diseases_lst_
    global log_file_
    global parameters_
    #assign values
    phenotype_df_exposed_ = phenotype_df_exposed
    id_col_ = id_col
    trajectory_ineligible_ = trajectory_ineligible
    trajectory_eligible_withdate_ = trajectory_eligible_withdate
    all_diagnosis_level_ = all_diagnosis_level
    covariates_ = covariates
    all_diseases_lst_ = all_diseases_lst
    log_file_ = log_file
    parameters_ = parameters
    # Call the original function
    return logistic_model((d1, d2))

def init_worker(phenotype_df_exposed:pd.DataFrame,id_col,trajectory_ineligible:dict,
                trajectory_eligible_withdate:dict,all_diagnosis_level:dict,covariates:list,
                all_diseases_lst:list,log_file:str,parameters:dict):
    """
    This function sets up the necessary global variables for a worker process in a multiprocessing environment.
    It assigns the provided parameters to global variables that can be accessed by logistic_model function in the worker process.

    Parameters:
    ----------
    phenotype_df_exposed : pd.DataFrame, phenotypic data for exposed individuals only.
    trajectory_ineligible : dict, trajectory ineligible disease dictionary.
    trajectory_eligible_withdate : dict, trajectory eligible disease (with date) dictionary.
    all_diagnosis_level : list, list of all diagnosed phecodes, with phecode truncated to corresponding level
    covariates : list, list of covariates to be included in the model.
    all_diseases_lst : list, list of other diseases to be included.
    log_file : str, Path and prefix for the log file
    parameters : dict, other arguments, including method and the associated parameters.

    Returns:
    ----------
    None

    """
    #shared global data
    global phenotype_df_exposed_
    global id_col_
    global trajectory_ineligible_
    global trajectory_eligible_withdate_
    global all_diagnosis_level_
    global covariates_
    global all_diseases_lst_
    global log_file_
    global parameters_
    #assign values
    phenotype_df_exposed_ = phenotype_df_exposed
    id_col_ = id_col
    trajectory_ineligible_ = trajectory_ineligible
    trajectory_eligible_withdate_ = trajectory_eligible_withdate
    all_diagnosis_level_ = all_diagnosis_level
    covariates_ = covariates
    all_diseases_lst_ = all_diseases_lst
    log_file_ = log_file
    parameters_ = parameters
    
    
    